local Boat = ...

--------------------------------------------------------------------------
local HARBOUR_COMP_BOAT = {
	TypeName = HarbourPrefix .. "_COMP_BOAT",
	ParentType = "COMPONENT",
	Properties = {
--		{ Name = "BoatOffsetY", Type = "float", Default = 2.11 },
		{ Name = "FlagTerrain", Type = "integer", Default = 2 ^ OBJECT_FLAG.TERRAIN:toNumber() + 2 ^ OBJECT_FLAG.BUILDING_PART:toNumber() + 2 ^ OBJECT_FLAG.PLATFORM:toNumber()},
                { Name = "BoatSpeed", Type = "float", Default = 1.0 },
                { Name = "BoatLength", Type = "float", Default = 10.0 },
                { Name = "RandomPath", Type = "boolean", Default = false },
                { Name = "MaxRandomDistanceMove", Type = "float", Default = 50.0 },
                { Name = "MaxBoatNameFlagCompNumber", Type = "integer", Default = 0 },
                { Name = "_onEnabledNumber", Type = "integer", Default = 0 },
	}
}
function HARBOUR_COMP_BOAT:registerAssignedSailor(_sailor)
    self._assignedSailor = _sailor
end

--------------------------------------------------------------------------
function HARBOUR_COMP_BOAT:create()
    self._boatSequenceNumberPath = 0
    self._assignedSailor = nil
    self._onEnabledNumber = 0
    self.MaxBoatNameFlagCompNumber = 2 -- don't currently works with 0
    self._initTimer = 10
end

--------------------------------------------------------------------------
function HARBOUR_COMP_BOAT:init()
        Boat:log("init: RandomPath = " .. tostring(self.RandomPath) .. tostring(self._assignedSailor))
        Boat:log("init: _boatSequenceNumberPath = " .. self._boatSequenceNumberPath)
        self._onEnabledNumber = 0
        self._initTimer = 10
end


--------------------------------------------------------------------------
function HARBOUR_COMP_BOAT:onDisabled()
        self._onEnabledNumber = self._onEnabledNumber - 1
end
--------------------------------------------------------------------------
function HARBOUR_COMP_BOAT:isSailorEmbarked()
        if self._assignedSailor == nil then
            Boat:log("onEnabled 0: RandomPath = " .. tostring(self.RandomPath) .. tostring(self._assignedSailor) .. self._onEnabledNumber)
            Boat:log("onEnabled 0: _boatSequenceNumberPath = " .. self._boatSequenceNumberPath)
            return false
        end
        if self._onEnabledNumber < 1 then
            Boat:log("onEnabled 0: RandomPath = " .. tostring(self.RandomPath) .. tostring(self._assignedSailor) .. self._onEnabledNumber)
            Boat:log("onEnabled 0: _boatSequenceNumberPath = " .. self._boatSequenceNumberPath)
 
            return false-- workaround awaiting to find the root cause of duplicated onEnabled
        end
        local _guardingLocationComp = self:getOwner():findFirstObjectWithComponentDown(HarbourPrefix .. "_COMP_GUARDING_LOCATION")
        if _guardingLocationComp == nil then       
            Boat:log("onEnabled: " .. HarbourPrefix .. "_COMP_GUARDING_LOCATION nil")
            return false
        end
        local deltaX = 	_guardingLocationComp:getOwner():getGlobalPosition()[1] - self._assignedSailor:getGlobalPosition()[1] 
        local deltaZ = 	_guardingLocationComp:getOwner():getGlobalPosition()[3] - self._assignedSailor:getGlobalPosition()[3] 
        local delta = deltaX*deltaX + deltaZ*deltaZ 
        if delta > 1 then
            Boat:log("onEnabled 0: RandomPath = " .. tostring(self.RandomPath) .. tostring(self._assignedSailor) .. self._onEnabledNumber .. " delta = " .. delta)
            Boat:log("onEnabled 0: _boatSequenceNumberPath = " .. self._boatSequenceNumberPath)
            return false
        end
        Boat:log("onEnabled: RandomPath = " .. tostring(self.RandomPath) .. tostring(self._assignedSailor) .. " delta = " .. delta)
        Boat:log("onEnabled: _boatSequenceNumberPath = " .. self._boatSequenceNumberPath)
        return true
end
--------------------------------------------------------------------------
function HARBOUR_COMP_BOAT:launchShip()
        if self.RandomPath then
            self:chooseRandomSequence()
        else
            self._initialPosition = self:getOwner():getGlobalPosition()
	    self:chooseNextFlag()
        end
        if self._boatSequenceNumberPath > 1 then
            self:getOwner():setRotationY(self._rotationY + math.pi/2)
        end
        Boat:log("FlagTerrain = " .. self.FlagTerrain .. " bit TERRAIN = " .. OBJECT_FLAG.TERRAIN:toNumber() .. " self._assignedSailor = " .. tostring(self._assignedSailor))
-- doesn't work
--        if self._assignedSailor ~= nil then
--           self._assignedSailor:setParent(self:getOwner())
--        end
        self.sailorDeltaY = self:getOwner():getGlobalPosition()[2] - self._assignedSailor:getGlobalPosition()[2]
        Boat:log("FlagTerrain = " .. self.FlagTerrain .. " bit TERRAIN = " .. OBJECT_FLAG.TERRAIN:toNumber() .. " self._assignedSailor = " .. tostring(self._assignedSailor))
end
--------------------------------------------------------------------------
function HARBOUR_COMP_BOAT:onEnabled()
        self._onEnabledNumber = self._onEnabledNumber + 1
        if self:isSailorEmbarked() then
            self:launchShip()
        end
end
--------------------------------------------------------------------------
function HARBOUR_COMP_BOAT:chooseRandomSequence()
        self._timer = math.random() * self.MaxRandomDistanceMove / self.BoatSpeed
        self._rotationY = (math.random() - 0.5) * 2 * math.pi
--        Boat:log("chooseRandomSequence: _rotationY = " .. tostring(self._rotationY))
        self._boatSequenceNumberPath = 2        
end

function HARBOUR_COMP_BOAT:updateBoatNextPosition(_initialSequence)
        Boat:log("updateBoatNextPosition: _initialSequence = " .. _initialSequence)
        local BoatNameFlagComp = HarbourPrefix .. "_COMP_FLAG"
        if _initialSequence < 3 then
            BoatNameFlagComp = HarbourPrefix .. "_COMP_FLAG1"
        else
            local BoatFlagNumber = _initialSequence - 1
            if BoatFlagNumber > self.MaxBoatNameFlagCompNumber then
                Boat:logWarning("No more path defined")
                return false
            end
            BoatNameFlagComp = HarbourPrefix .. "_COMP_FLAG" .. BoatFlagNumber 
        end
        Boat:log("BoatNameFlagComp = " .. BoatNameFlagComp)

        Boat:log("GlobalPosition = " .. tostring(self:getOwner():getGlobalPosition()))

        local _nextFlag = self:getLevel():find(BoatNameFlagComp, false)        
        if _nextFlag == nil then
            Boat:logWarning("No more path defined")
            return false
        end
        Boat:log("_nextFlag = " .. tostring(_nextFlag))
        if not _nextFlag.BoatIsFlagSet then
            Boat:logWarning("Flag not set")
            return false
        end
        if _nextFlag:getOwner() == nil then    
            Boat:logWarning("No flag owner defined")
            return false
        end
        Boat:log("_nextFlag:getOwner() = " .. tostring(_nextFlag:getOwner()))
        local _flagObject = _nextFlag:getOwner()
        Boat:log("_positionFlag = " .. tostring(_flagObject:getGlobalPosition()))
        self._nextPosition = _flagObject:getGlobalPosition()       
        Boat:log("_nextPosition = " .. tostring(self._nextPosition))
        return true
end

function HARBOUR_COMP_BOAT:updateBoatSequenceNumberPath(_initialSequence, _isNewPosition)
        Boat:log("_initialSequence = " .. _initialSequence .. " _isNewPosition = " .. tostring(_isNewPosition))
        if _isNewPosition then
            if _initialSequence < 3 then
                Boat:log("Step 3: move to next position")
                return 3
            else 
                Boat:log("Step  greater than 3: move to next position")
                return _initialSequence + 1
            end
        else
            if _initialSequence < 3 then

                Boat:log("Step 1: nothing to do")
                return 1 -- nothing to do
            else
                self._nextPosition = self._initialPosition
                Boat:log("Step 2: return to initial position")
                return 2 -- return to initial position
            end
        end
end

function HARBOUR_COMP_BOAT:updateBoatNextMove()
                local DeltaX = self._nextPosition[1] - self:getOwner():getGlobalPosition()[1]
                local DeltaZ = self._nextPosition[3] - self:getOwner():getGlobalPosition()[3]
                local Distance = (DeltaX*DeltaX + DeltaZ*DeltaZ) ^ 0.5
                Boat:log("angle : X = " .. DeltaX .. " Z = " .. DeltaZ .. " Distance = " .. Distance)
                if Distance > 0 then
                    if DeltaX > Distance then
                        DeltaX = 1.0
                    elseif DeltaX < -Distance then
                        DeltaX = -1.0
                    else
                        DeltaX = DeltaX / Distance
                    end 
                    self._angleX = math.asin(DeltaX)
                    if DeltaZ > Distance then
                        DeltaZ = 1.0
                    elseif DeltaZ < -Distance then
                        DeltaZ = -1.0
                    else
                        DeltaZ = DeltaZ / Distance
                    end 
                    if DeltaZ > 0 then
                        self._angleZ = self._angleX
                    else
                        self._angleZ = math.pi - self._angleX
                    end
                    self._rotationY = self._angleZ                  
                    Boat:log("self._rotationY = " .. self._rotationY .. " X = " .. self._angleX .. " Z = " .. self._angleZ)
                    self._timer = Distance / self.BoatSpeed
                end 
end

--------------------------------------------------------------------------
function HARBOUR_COMP_BOAT:chooseNextFlag()
        local IsUpdatedBoatNextPosition = false
        Boat:log("IsUpdatedBoatNextPosition before = " .. tostring(IsUpdatedBoatNextPosition))
        IsUpdatedBoatNextPosition = self:updateBoatNextPosition(self._boatSequenceNumberPath)
        Boat:log("IsUpdatedBoatNextPosition after = " .. tostring(IsUpdatedBoatNextPosition))
        self._boatSequenceNumberPath = self:updateBoatSequenceNumberPath(self._boatSequenceNumberPath, IsUpdatedBoatNextPosition)
        Boat:log("Sequence after = " .. self._boatSequenceNumberPath)

        self._timer = 0.0
        self._rotationY = 0
        if self._boatSequenceNumberPath > 1 then
            self:updateBoatNextMove()
            Boat:log("self._timer = " .. self._timer)
        end
end

--------------------------------------------------------------------------
function HARBOUR_COMP_BOAT:checkWaterElevation(_boatDestination)
    local raycastResult = {}
--    Boat:log("initial raycastResult = " .. tostring(raycastResult))
     
    -- Raycast from the building position + Y = 100
    -- to building position - y = 100
    -- only on objects with a WATER flag
    local FromPosition = { _boatDestination[1], _boatDestination[2]+100, _boatDestination[3] }
    local ToPosition = { _boatDestination[1], _boatDestination[2]-100, _boatDestination[3] }
--    Boat:log("FromPosition = " .. tostring(FromPosition) .. "ToPosition = " .. tostring(ToPosition))
    
--    Boat:log("final raycastResult = RigidBody = " .. tostring(raycastResult["RigidBody"]) .. " Position = " .. tostring(raycastResult["Position"]) .. " Normal = " .. tostring(raycastResult["Normal"]))
    if not self:getLevel():rayCast(FromPosition, ToPosition, raycastResult, 2 ^ OBJECT_FLAG.WATER:toNumber()) then
         Boat:log("FromPosition = " .. tostring(FromPosition) .. "ToPosition = " .. tostring(ToPosition))
         self._timer = -100
         Boat:logWarning("Water Mill Set: Boat NOT on water (out of map ?)")
         return 0
    else
--        Boat:log("checkWaterElevation = " .. - _boatDestination[2] + raycastResult["Position"][2])
         return -_boatDestination[2] + raycastResult["Position"][2]
    end
end

--------------------------------------------------------------------------
function HARBOUR_COMP_BOAT:update()
--        Boat:log("update: RandomPath = " .. tostring(self.RandomPath))
--        Boat:log("update: _boatSequenceNumberPath = " .. self._boatSequenceNumberPath)
    local dt = self:getLevel():getDeltaTime()
    if self._boatSequenceNumberPath == 0 then
        -- case of reload
        if self._onEnabledNumber == 0 then
            self._initTimer = 10
            return
        end
        if not self:isSailorEmbarked() then
            self._initTimer = 10
            return
        end
        if self._initTimer >= 0 then
            self._initTimer = self._initTimer - dt
            return
        end
        self:launchShip()
        return
    end
    if self._boatSequenceNumberPath > 0 then
        dt = self:getLevel():getDeltaTime()
        if self._timer < 0 then

            if self.RandomPath then
                self:chooseRandomSequence()
            else
                self:chooseNextFlag()
            end          
            if self._boatSequenceNumberPath > 1 then
                self:getOwner():setRotationY(self._rotationY + math.pi/2)
-- to be enhanced: the rotation should be smoothed and the sailor should follow the move if the GuardLocation is not close to the pivot of the ship
                if self._assignedSailor ~= nil then
                    self._assignedSailor:setGlobalPosition({self:getOwner():getGlobalPosition()[1], self:getOwner():getGlobalPosition()[2], self:getOwner():getGlobalPosition()[3]}) --security to avoid the "drifts"
                end
            end
       end
       self._timer = self._timer - dt
--       Boat:log("update: _timer = " .. tostring(self._timer))
    end

    if self._boatSequenceNumberPath > 1 then
        local BoatCurrent = self:getOwner():getGlobalPosition()
--        Boat:log("BoatCurrent = " .. tostring(BoatCurrent))
        local raycastResult = {}
--        Boat:log("initial raycastResult = " .. tostring(raycastResult))
     
    -- Raycast from the last current position to the next position + BoatLength
    -- only on objects with a TERRAIN flag
        local q = self:getOwner():getGlobalOrientation()
--        Boat:log("Boat:q = " .. tostring(q) ) 
        local sinp = 2 * q[4] * q[2]
        local cosp = q[4] * q[4] - q[2] * q[2]
        local normep = cosp*cosp + sinp*sinp
        if (normep < 0.9) or (normep > 1.1)  then
            Boat:logError("getGlobalOrientation() error: " .. tostring(q))
        end
--        Boat:log("sinp =" .. sinp .. " cosp = " .. cosp)
        local BoatDestination = { self:getOwner():getGlobalPosition()[1] - cosp * self.BoatLength, 
                                  self:getOwner():getGlobalPosition()[2],
                                  self:getOwner():getGlobalPosition()[3] + sinp * self.BoatLength}
--        Boat:log("BoatDestination = " .. tostring(BoatDestination))
        if self:getLevel():rayCast(self:getOwner():getGlobalPosition(), BoatDestination, raycastResult, self.FlagTerrain) then
--            Boat:log("collision with BUILDING_PART raycastResult = RigidBody = " .. tostring(raycastResult["RigidBody"]) .. " Position = " .. tostring(raycastResult["Position"]) .. " Normal = " .. tostring(raycastResult["Normal"]))
            self._timer = -1 -- TERRAIN Reached, force new path
            return
        end
--        Boat:log("final raycastResult = RigidBody = " .. tostring(raycastResult) .. tostring(raycastResult["RigidBody"]) .. " Position = " .. tostring(raycastResult["Position"]) .. " Normal = " .. tostring(raycastResult["Normal"]))
    
        local BoatEndDestination = { self:getOwner():getGlobalPosition()[1] - cosp * dt * self.BoatSpeed, 
                                     self:getOwner():getGlobalPosition()[2],
                                     self:getOwner():getGlobalPosition()[3] + sinp * dt * self.BoatSpeed}
        self:getOwner():translate({ math.sin(self._rotationY) * dt * self.BoatSpeed, self:checkWaterElevation(BoatEndDestination), math.cos(self._rotationY) * dt * self.BoatSpeed })
        if self._assignedSailor ~= nil then
            local q = self._assignedSailor:getGlobalOrientation()
--        Boat:log("Boat:q = " .. tostring(q) ) 
            local sins = 2 * q[4] * q[2]
            local coss = q[4] * q[4] - q[2] * q[2]
            local normes = coss*coss + sins*sins
            if (normes < 0.9) or (normes > 1.1)  then
                Boat:logError("getGlobalOrientation() error: " .. tostring(q))
            end
            self._assignedSailor:translate({ (coss*math.sin(self._rotationY) + sins*math.cos(self._rotationY)) * dt * self.BoatSpeed, 
                                             self:checkWaterElevation(BoatEndDestination) - self.sailorDeltaY,
                                             (coss*math.cos(self._rotationY) - sins*math.sin(self._rotationY)) * dt * self.BoatSpeed })
        end
        if self._timer == -100 then --error detected by checkWaterElevation
            self:getOwner():destroy()
        end 
     end 

end

--------------------------------------------------------------------------
Boat:registerClass(HARBOUR_COMP_BOAT)
